@@ -1,11 +1,11 @@
 package Helios::Service;
 
-use 5.008000;
+use 5.008;
 use base qw( TheSchwartz::Worker );
 use strict;
 use warnings;
+use File::Spec;
 use Sys::Hostname;
-use Sys::Syslog qw(:standard :macros);
 use Config::IniFiles;
 use DBI;
 use Data::ObjectDriver::Driver::DBI;
@@ -19,19 +19,24 @@ use Helios::Error;
 use Helios::Job;
 use Helios::ConfigParam;
 use Helios::LogEntry;
+use Helios::LogEntry::Levels qw(:all);
 
-our $VERSION = '2.00';
-our $MAX_RETRIES = 0;
-our $RETRY_INTERVAL = 0;
+our $VERSION = '2.30_4931';
+
+our $CACHED_CONFIG;
+our $CACHED_CONFIG_RETRIEVAL_COUNT = 0;
+our $WORKER_START_TIME = 0;
+
+our %INIT_LOG_CLASSES;	# for the logging system
 
 
 =head1 NAME
 
-Helios::Service - base class for services/worker classes in the Helios job processing system
+Helios::Service - base class for services in the Helios job processing system
 
 =head1 DESCRIPTION
 
-Helios::Service is the base class for all services (aka worker classes) intended to be run by the 
+Helios::Service is the base class for all services intended to be run by the 
 Helios parallel job processing system.  It handles the underlying TheSchwartz job queue system and 
 provides additional methods to handle configuration, job argument parsing, logging, and other 
 functions.
@@ -43,8 +48,14 @@ failedJob(), or failedJobPermanent(), respectively) before it ends.
 
 =head1 TheSchwartz HANDLING METHODS
 
-The following 3 methods are used by the underlying TheSchwartz job queuing system to determine what
-work is to be performed and, a job fails, how it should be retried.
+The following 3 methods are used by the underlying TheSchwartz job queuing 
+system to determine what work is to be performed and, a job fails, how it 
+should be retried.
+
+YOU DO NOT NEED TO TOUCH THESE METHODS TO CREATE HELIOS SERVICES.  These 
+methods manage interaction between Helios and TheSchwartz.  You only need to 
+be concerned with these methods if you are attempting to extend core Helios
+functionality.  
 
 =head2 max_retries()
 
@@ -80,24 +91,47 @@ true value, the worker process will exit.  If OVERDRIVE mode is disabled, the pr
 
 =cut
 
+sub max_retries { return MaxRetries(); }
+sub retry_delay { return RetryInterval(); }
+
 sub work {
 	my $class = shift;
 	my $schwartz_job = shift;
-	my $job = Helios::Job->new($schwartz_job);
+	my $job = $class->JobClass() ? $class->JobClass()->new($schwartz_job) : Helios::Job->new($schwartz_job);
+	$WORKER_START_TIME = $WORKER_START_TIME ? $WORKER_START_TIME : time();     # for WORKER_MAX_TTL 
 	my $return_code;
 	my $args;
 
 	# instantiate the service class into a worker
 	my $self = new $class;
-	$MAX_RETRIES = $self->max_retries();
-	$RETRY_INTERVAL = $self->retry_delay();
-	if ($self->debug) {
-		print "max_retries: ",max_retries(),"\n";
-		print "retry_delay: ",retry_delay(),"\n";
-	}
-
 	try {
-		$self->prep();
+	    # if we've previously retrieved a config
+        # AND OVERDRIVE is enabled (1) 
+        # AND LAZY_CONFIG_UPDATE is enabled (1),
+        # AND we're not servicing the 10th job (or technically a multiple of ten)
+        # THEN just retrieve the pre-existing config
+        
+        #[]
+        print "CACHED_CONFIG=",$CACHED_CONFIG,"\n";
+        print "CACHED_CONFIG_RETRIEVAL_COUNT=",$CACHED_CONFIG_RETRIEVAL_COUNT,"\n";
+        
+        if ( defined($CACHED_CONFIG) && 
+                $CACHED_CONFIG->{LAZY_CONFIG_UPDATE} == 1 &&
+                $CACHED_CONFIG->{OVERDRIVE} == 1 &&
+                $CACHED_CONFIG_RETRIEVAL_COUNT % 10 != 0 
+            ) {
+            $self->prep($CACHED_CONFIG);
+            $CACHED_CONFIG_RETRIEVAL_COUNT++;
+            if ($self->debug) { $self->logMsg(LOG_DEBUG,"Retrieved config params from in-memory cache"); }    #[] take this out before release
+        } else {
+            $self->prep();
+            if ( defined($self->getConfig()->{LAZY_CONFIG_UPDATE}) && 
+                    $self->getConfig()->{LAZY_CONFIG_UPDATE} == 1 ) {
+                $CACHED_CONFIG = $self->getConfig();
+                $CACHED_CONFIG_RETRIEVAL_COUNT = 1;     # "prime the pump"
+            }	    
+        }
+	    	    
 		$job->setConfig($self->getConfig());
 		$job->debug( $self->debug );
 		$args = $job->parseArgs();
@@ -120,11 +154,12 @@ sub work {
 
 	# if this is a meta job, we need to burst it apart 
 	# if it isn't, we should call the service's run() method 
+	$self->setJob($job);
 	if ( defined($args->{metajob}) && $args->{metajob} == 1 ) {
-		$self->logMsg($job, LOG_INFO, 'Job '.$job->getJobid.' is a metajob.  Bursting...');
+		$self->logMsg($job, LOG_NOTICE, 'Job '.$job->getJobid.' is a metajob.  Bursting...');
 		try {
 			my $numberOfJobs = $self->burstJob($job);
-			$self->logMsg($job, LOG_INFO, 'Job '.$job->getJobid.' burst into '.$numberOfJobs.' jobs');
+			$self->logMsg($job, LOG_NOTICE, 'Job '.$job->getJobid.' burst into '.$numberOfJobs.' jobs');
 		} otherwise {
 			my $e = shift;
 			$self->logMsg($job, LOG_ERR, 'Job '.$job->getJobid.' burst failure: '.$e->text);
@@ -147,7 +182,7 @@ sub work {
 	# if we're in OVERDRIVE, work() will exit and the worker process will call it again with another job
 	# if we were in OVERDRIVE, but now we're NOT, we should explicitly exit() to accomplish the downshift
 	if ( $self->shouldExitOverdrive() ) {
-		$self->logMsg(LOG_INFO,"Class $class exited (downshift)");
+		$self->logMsg(LOG_NOTICE,"Class $class exited (downshift)");
 		exit(0);
 	}
 }
@@ -160,6 +195,7 @@ These accessors will be needed by subclasses of Helios::Service.
  get/setConfig()
  get/setHostname()
  get/setIniFile()
+ get/setJob()
  get/setJobType()
  errstr()
  debug()
@@ -176,6 +212,9 @@ also use it within your own service classes to enable/disable debugging messages
 
 =cut
 
+sub setJob { $_[0]->{job} = $_[1]; }
+sub getJob { return $_[0]->{job}; }
+
 # need for helios.pl logging	
 sub setJobType { $_[0]->{jobType} = $_[1]; }
 sub getJobType { return $_[0]->{jobType}; }
@@ -183,6 +222,9 @@ sub getJobType { return $_[0]->{jobType}; }
 sub setConfig { $_[0]->{config} = $_[1]; }
 sub getConfig { return $_[0]->{config}; }
 
+sub setFuncid { $_[0]->{funcid} = $_[1]; }
+sub getFuncid { return $_[0]->{funcid}; }
+
 sub setIniFile { $_[0]->{inifile} = $_[1]; }
 sub getIniFile { return $_[0]->{inifile}; }
 
@@ -200,7 +242,7 @@ sub debug { my $self = shift; @_ ? $self->{debug} = shift : $self->{debug}; }
 The new() method doesn't really do much except create an object of the appropriate class.  (It can 
 overridden, of course.)
 
-It does set the job type for the object (available via the getJobType() method)
+It does set the job type for the object (available via the getJobType() method).
 
 =cut
 
@@ -221,10 +263,16 @@ sub new {
 
 =head1 INTERNAL SERVICE CLASS METHODS
 
+When writing normal Helios services, the methods listed in this section will 
+have already been dealt with before your run() method is called.  If you are 
+extending Helios itself or instantiating a Helios service outside of Helios 
+(for example, to retrieve a service's config params), you may be interested in 
+some of these, primarily the prep() method. 
+
 =head2 prep()
 
 The prep() method is designed to call all the various setup routines needed to 
-get the worker ready to do useful work.  It:
+get the service ready to do useful work.  It:
 
 =over 4
 
@@ -252,6 +300,7 @@ an exception, that exception will be raised to your calling routine.
 
 sub prep {
 	my $self = shift;
+	my $cached_config = shift;
 
 	# pull params from environment
 	$self->setHostname(hostname);
@@ -263,11 +312,17 @@ sub prep {
 		$self->setIniFile($ENV{HELIOS_INI});
 	}
 
+    if ( defined($cached_config) ) {
+        $self->setConfig($cached_config);
+        return 1;        
+    }
+
 	# now get the Helios conf params from INI and db
 	# (these may throw their own errors that the calling routine
 	#  will have to catch)
 	$self->getConfigFromIni();
 	$self->getConfigFromDb();
+
 	return 1;
 }
 
@@ -278,7 +333,7 @@ The getConfigFromIni() method opens the helios.ini file, grabs global params and
 the current service class, and returns them in a hash to the calling routine.  It also sets the class's 
 internal {config} hashref, so the config parameters are available via the getConfig() method.
 
-Typically job service classes will call this once near the start of processing to pick up any relevant 
+Typically service classes will call this once near the start of processing to pick up any relevant 
 parameters from the helios.ini file.  However, calling the prep() method takes care of this for 
 you, and is the preferred method.
 
@@ -297,8 +352,8 @@ sub getConfigFromIni {
 			$inifile = $self->getIniFile();
 		} elsif ( defined($ENV{HELIOS_INI}) ) {
 			$inifile = $ENV{HELIOS_INI};
-		} elsif (-r './helios.ini' ) {
-			$inifile = './helios.ini';
+		} elsif (-r File::Spec->catfile(File::Spec->curdir,'helios.ini') ) {
+			$inifile = File::Spec->catfile(File::Spec->curdir,'helios.ini');
 		} else {
 			throw Helios::Error::InvalidArg("INI configuration file not specified.");
 		}
@@ -319,7 +374,7 @@ sub getConfigFromIni {
 		throw Helios::Error::InvalidArg("Section [global] doesn't exist in config file $inifile");
 	}
 	
-	# if there's a section specifically for this worker class, read it too
+	# if there's a section specifically for this service class, read it too
 	# (it will effectively override the global section, BTW)
 	if ( $ini->SectionExists($jobtype) ) {
 		foreach ( $ini->Parameters($jobtype) ) {
@@ -336,7 +391,7 @@ sub getConfigFromIni {
 
 The getConfigFromDb() method connects to the Helios database, retrieves config params relevant to the 
 current service class, and returns them in a hash to the calling routine.  It also sets the class's 
-internal {config} hasref, so the config parameters are available via the getConfig() method.
+internal {config} hashref, so the config parameters are available via the getConfig() method.
 
 Typically service classes will call this once near the start of processing to pick up any relevant 
 parameters from the helios.ini file.  However, calling the prep() method takes care of this for 
@@ -385,7 +440,7 @@ sub getConfigFromDb {
 			}
 			$params->{$cp->param()} = $cp->value();
 		}
-
+		
 	} otherwise {
 		my $e = shift;
 		throw Helios::Error::DatabaseError($e->text);
@@ -396,10 +451,41 @@ sub getConfigFromDb {
 }
 
 
+=head2 getFuncidFromDb()
+
+=cut
+
+sub getFuncidFromDb {
+    my $self = shift;
+    my $params = $self->getConfig();
+    my $jobtype = $self->getJobType();
+    my @funcids;
+
+    if ($self->debug) { print "Retrieving funcid for ".$self->getJobType()."\n"; }
+
+    try {
+        my $driver = $self->getDriver();
+        # also get the funcid 
+        my @funcids = $driver->search('TheSchwartz::FuncMap' => {
+                funcname => $jobtype 
+            }
+        );
+        if ( scalar(@funcids) > 0 ) {
+            $self->setFuncid( $funcids[0]->funcid() );
+        }
+		
+	} otherwise {
+		my $e = shift;
+		throw Helios::Error::DatabaseError($e->text);
+	};
+    return $self->getFuncid();	
+}
+
+
 =head2 jobsWaiting() 
 
-Scans the job queue for jobs that are ready to run.  Returns the number of jobs waiting.  Only 
-meant for use with the helios.pl service daemon.
+Scans the job queue for jobs that are ready to run.  Returns the number of jobs 
+waiting.  Only meant for use with the helios.pl service daemon.
 
 =cut
 
@@ -407,22 +493,19 @@ sub jobsWaiting {
 	my $self = shift;
 	my $params = $self->getConfig();
 	my $jobType = $self->getJobType();
+	my $funcid;
 
 	try {
 
 		my $dbh = $self->dbConnect($params->{dsn}, $params->{user}, $params->{password});
 		unless ($dbh) { throw Helios::Error::DatabaseError($self->errstr); }
 
-		# get the funcid 
-		my $sth1 = $dbh->prepare("SELECT funcid FROM funcmap WHERE funcname = ?");
-
-		$sth1->execute($jobType);
-
-		my $result1 = $sth1->fetchrow_arrayref();
-		unless ( defined($result1->[0]) ) { 
-			$self->errstr("funcname $jobType not found!"); return undef; 
+		# get the funcid if we don't already have it
+		if ( !defined($self->getFuncid()) ) {
+		    $funcid = $self->getFuncidFromDb();
+		} else {
+		    $funcid = $self->getFuncid();
 		}
-		my $funcid = $result1->[0];
 
 		my $sql2 = <<JWSQL2;
 SELECT COUNT(*)
@@ -434,12 +517,11 @@ JWSQL2
 		my $sth2 = $dbh->prepare($sql2);
 
 		my $current_time = time();
-		$sth2->execute($result1->[0], $current_time, $current_time);
+		$sth2->execute($funcid, $current_time, $current_time);
 		my $result2 = $sth2->fetchrow_arrayref();
 		unless ( defined($result2->[0]) ) {
 			$self->errstr("Received NULL value in jobsWaiting() for funcname $jobType!"); return undef; 
 		}
-		$sth1->finish();
 		$sth2->finish();
 		$dbh->disconnect();
 		return $result2->[0];
@@ -471,9 +553,13 @@ sub getDriver {
 
 =head2 shouldExitOverdrive()
 
-Determine whether or not to exit if OVERDRIVE mode is enabled.  The config params will be checked 
-for HOLD, HALT, or OVERDRIVE values.  If HALT is defined or HOLD == 1 this method will return a 
-true value, indicating the worker process should exit().
+Determine whether or not to exit if OVERDRIVE mode is enabled.  The config 
+params will be checked for HOLD, HALT, or OVERDRIVE values.  If HALT is defined 
+or HOLD == 1 this method will return a true value, indicating the worker 
+process should exit().
+
+This method is used by helios.pl and Helios::Service->work().  Normal Helios
+services do not need to use this method directly.
 
 =cut
 
@@ -482,12 +568,19 @@ sub shouldExitOverdrive {
 	my $params = $self->getConfig();
 	if ( defined($params->{HALT}) ) { return 1; }
 	if ( defined($params->{HOLD}) && $params->{HOLD} == 1) { return 1; }
+	if ( defined($params->{WORKER_MAX_TTL}) && $params->{WORKER_MAX_TTL} > 0 && 
+	       time() > $WORKER_START_TIME + $params->{WORKER_MAX_TTL} ) {
+        return 1;
+    }
 	return 0;
 }
 
 
 =head1 METHODS AVAILABLE TO SERVICE SUBCLASSES
 
+The methods in this section are available for use by Helios services.  They 
+allow your service to interact with the Helios environment.
+
 =head2 dbConnect($dsn, $user, $password)
 
 Method to connect to a database.  If parameters not specified, uses dsn, user, password 
@@ -526,7 +619,8 @@ sub dbConnect {
 			if ($self->debug) { print "dsn=$dsn\nuser=$user\npass=$password\n";	} 
 			$dbh = DBI->connect_cached($dsn, $user, $password);
 		}
-		if ( $DBI::errstr ) { 
+#[]		if ( $DBI::errstr ) { 
+        unless ( defined($dbh) ) {
 			$self->errstr("DB ERROR: ".$DBI::errstr); 
 			throw Helios::Error::DatabaseError($DBI::errstr);
 		}
@@ -540,10 +634,12 @@ sub dbConnect {
 }
 
 
-=head2 logMsg([$job,] [$priority,] $msg)
+=head2 logMsg([$job,] [$priority_level,] $message)
 
-Record a message in the Helios log.  If a syslogd log facility is set in helios.ini (via the 
-syslog_facility option), the message is also logged into the specified syslogd facility.
+Given a message to log, an optional priority level, and an optional Helios::Job
+object, logMsg() will record the message in the logging systems that have been 
+configured.  The internal Helios logging system is the only system enabled by 
+default.
 
 In addition to the log message, there are two optional parameters:
 
@@ -551,147 +647,185 @@ In addition to the log message, there are two optional parameters:
 
 =item $job
 
-The current Helios::Job object being processed.  If specified, the jobid will be logged in the 
-database along with the message.
+The current Helios::Job object being processed.  If specified, the jobid will 
+be logged in the database along with the message.
 
 =item $priority
 
-The priority of the message as defined by syslogd.  These are really integers, but if you import 
-the syslog constants [use Sys::Syslog qw(:macros)] into your namespace, your logMsg() calls will 
-be much more readable.  Refer to the L<Sys::Syslog/CONSTANTS> manpage for a list of valid syslog
-constants.  LOG_DEBUG, LOG_INFO, LOG_WARNING, and LOG_ERR are the most common used with Helios; 
-LOG_INFO is the default.
+The priority level of the message as defined by Helios::LogEntry::Levels.  
+These are really integers, but if you import Helios::LogEntry::Levels (with the 
+:all tag) into your namespace, your logMsg() calls will be much more readable.  
+There are 8 log priority levels, corresponding (for historical reasons) to 
+the log priorities defined by Sys::Syslog:
+
+    name         priority
+    LOG_EMERG    0
+    LOG_ALERT    1
+    LOG_CRIT     2
+    LOG_ERR      3
+    LOG_WARNING  4
+    LOG_NOTICE   5
+    LOG_INFO     6
+    LOG_DEBUG    7
+   
+LOG_DEBUG, LOG_INFO, LOG_NOTICE, LOG_WARNING, and LOG_ERR are the most common 
+used by Helios; LOG_INFO is the default.
 
 =back
 
-In addition, there are two INI file options used to configure logging.  These will be passed to 
-syslogd when logMsg() calls Sys::Syslog::openlog():
+The host, process id, and service class are automatically recorded with your log 
+message.  If you supplied either a Helios::Job object or a priority level, these
+will also be recorded with your log message.
 
-=over 4
+This method returns a true value if successful and throws an exception if errors occur.   
 
-=item syslog_facility
+=head3 LOGGING SYSTEM CONFIGURATION
 
-The syslog facility to log the message to.
+Several parameters are available to configure Helios logging.  Though these 
+options can be set either in helios.ini or in the Ctrl Panel, it is B<strongly>
+recommended these options only be set in helios.ini.  Changing logging 
+configurations on-the-fly could potentially cause a Helios service (and 
+possibly your whole collective) to become unstable!
 
-=item syslog_options
+The following options can be set in either a [global] section or in an 
+application section of your helios.ini file.
 
-Any logging options to specify to syslogd.  Again, see the L<Sys::Syslog> manpage.
+=head4 loggers
 
-=back
+ loggers=HeliosX::Logger::Syslog,HeliosX::Logger::Log4perl
 
-For database logging, the host, process id, and job service class are automatically recorded in 
-the database with your log message.  If you supplied either a Job object or a priority, 
-the jobid and/or priority will also be recorded with your log message.
+A comma delimited list of interface classes to external logging systems.  Each 
+of these classes should implement (or otherwise extend) the Helios::Logger 
+class.  Each class will most likely have its own configuration parameters to 
+set; consult the documentation for the interface class you're trying to 
+configure.
+
+=head4 internal_logger 
+
+ internal_logger=on|off 
+
+Whether to enable the internal Helios logging system as well as the loggers 
+specified with the 'loggers=' line above.  The default is on.  If set to off, 
+the only logging your service will do will be to the external logging systems.
+
+=head4 log_priority_threshold
+
+ log_priority_threshold=1|2|3|4|5|6   
+
+You can specify a logging threshold to better control the 
+logging of your service on-the-fly.  Unlike the above parameters, 
+log_priority_threshold can be safely specified in your Helios Ctrl Panel.  
+Specifying a 'log_priority_threshold' config parameter in your helios.ini or 
+Ctrl Panel will cause log messages of a lower priority (higher numeric value) 
+to be discarded.  For example, a line in your helios.ini like:
+
+ log_priority_threshold=6
+
+will cause any log messages of priority 7 (LOG_DEBUG) to be discarded.
+
+=head3 COPYRIGHT AND LICENSE
+
+The logMsg() method is Copyright (C) 2009 by Andrew Johnson
+
+This method is free software; you can redistribute it and/or modify
+it under the same terms as Perl itself, either Perl version 5.8.0 or,
+at your option, any later version of Perl 5 you may have available.
+
+=head3 WARRANTY
+
+This software comes with no warranty of any kind.
 
 =cut
 
 sub logMsg {
 	my $self = shift;
-	my $priority;
-	my $msg;
+	my @args = @_;
 	my $job;
+	my $level;
+	my $msg;
+	my @loggers;
 
-	# if the first parameter is a Helios::Job object, log extra info
-	if ( ref($_[0]) && $_[0]->isa('Helios::Job') ) {
-		$job = shift;
+	# were we called with 3 params?  ($job, $level, $msg)
+	# 2 params?                      ($level, $msg) or ($job, $msg)
+	# or just 1?                     ($msg)
+
+	# is the first arg is a Helios::Job object?
+	if ( ref($args[0]) && $args[0]->isa('Helios::Job') ) {
+		$job = shift @args;
 	}
 
-	# if we were given 2 params, the first is priority, second message
-	# if only one, it is the message, default to LOG_INFO priority
-	if ( defined($_[0]) && defined($_[1]) ) {
-		$priority = shift;
-		$msg = shift;
+	# if there are 2 params remaining, the first is level, second msg
+	# if only one, it's just the message 
+	if ( defined($args[0]) && defined($args[1]) ) {
+		$level = $args[0];
+		$msg = $args[1];
 	} else {
-		$priority = LOG_INFO;
-		$msg = shift;
+		$level = LOG_INFO;	# default the level to LOG_INFO
+		$msg = $args[0];
 	}
 	
-	my $params = $self->getConfig();
+	my $config = $self->getConfig();
 	my $jobType = $self->getJobType();
 	my $hostname = $self->getHostname();
 
-	# log the message to syslog IF log_facility was set
-	if ( defined($params->{syslog_facility}) ) {
-		openlog($jobType, $params->{syslog_options}, $params->{syslog_facility});
-		syslog($priority, $msg);
-		closelog();
-	}
+    # grab the names of all the configured loggers to try
+    @loggers = split(/,/, $config->{loggers});
 
-	# log to database
-	my $retries = 0;
-	my $retry_limit = 3;
-	RETRY:
-		try{
-			my $driver = $self->getDriver();
-			my $log_entry;
-			if ( defined($job) ) {
-				$log_entry = Helios::LogEntry->new(
-					log_time   => time(),
-					host       => $self->getHostname,
-					process_id => $$,
-					jobid      => $job->getJobid,
-					funcid     => $job->getFuncid,
-					job_class  => $jobType,
-					priority   => $priority,
-					message    => $msg
-				);
-			} else {
-				$log_entry = Helios::LogEntry->new(
-					log_time   => time(),
-					host       => $self->getHostname,
-					process_id => $$,
-					jobid      => undef,
-					funcid     => undef,
-					job_class  => $jobType,
-					priority   => $priority,
-					message    => $msg
-				);
+    # inject the internal logger automatically
+    # UNLESS it has been specifically turned off
+    unless ( defined($config->{internal_logger}) && 
+        ( $config->{internal_logger} eq 'off' || $config->{internal_logger} eq '0') ) {
+    	push(@loggers, 'Helios::Logger::Internal');
+    }
+
+	# if no loggers configured, this whole section will be skipped
+	foreach my $logger (@loggers) {
+		# init the logger if it hasn't been initialized yet
+		unless ( defined($INIT_LOG_CLASSES{$logger}) ) {
+			# attempt to init the class
+			unless ( $logger->can('init') ) {
+		        eval "require $logger";
+		        throw Helios::Error::LoggingError($@) if $@;
 			}
-			$driver->insert($log_entry);		
-
+			$logger->setConfig($config);
+			$logger->setJobType($jobType);
+			$logger->setHostname($hostname);
+            try {
+    			$logger->init();
+            } otherwise {
+            	# our only resort is to use the internal logger
+            	my $e = shift;
+            	print $e->text(),"\n"; #[]
+                Helios::Logger::Internal->setConfig($config);
+                Helios::Logger::Internal->setJobType($jobType);
+                Helios::Logger::Internal->setHostname($hostname);
+                Helios::Logger::Internal->init();
+            	Helios::Logger::Internal->logMsg(undef, LOG_EMERG, $logger.' CONFIGURATION ERROR: '.$e->text());
+            };
+			$INIT_LOG_CLASSES{$logger} = $logger;
+		}
+		try {
+		  $logger->logMsg($job, $level, $msg);
 		} otherwise {
-			my $e = shift;
-			if ($retries > $retry_limit) {
-				throw Helios::Error::DatabaseError($e->text());
-			} else {
-				# we're going to give it another shot
-				$retries++;
-				sleep 5;
-				next RETRY;
-			}
-		};
-	# retry block end
-	return 1;
-}
-
-
-=head2 parseArgXML($xml) [DEPRECATED]
-
-Given a string of XML, parse it into a mixed hash/arrayref structure.  This uses XML::Simple.
-
-This method is DEPRECATED; its functionality has been moved to Helios::Job->parseArgXML().
-
-=cut
-
-sub parseArgXML {
-	my $self = shift;
-	my $xml = shift;
-
-	my $xs = XML::Simple->new(SuppressEmpty => undef);
-	my $args;
-	try {
-		$args = $xs->XMLin($xml);
-	} otherwise {
-		throw Helios::Error::InvalidArg($!);
-	};
-	return $args;
+            my $e = shift;
+            print $e->text(),"\n"; #[]
+            Helios::Logger::Internal->setConfig($config);
+            Helios::Logger::Internal->setJobType($jobType);
+            Helios::Logger::Internal->setHostname($hostname);
+            Helios::Logger::Internal->init();
+            Helios::Logger::Internal->logMsg(undef, LOG_EMERG, $logger.' LOGGING FAILURE: '.$e->text());
+		};			
+	}
+	
+	return 1;	
 }
 
 
 =head2 getJobArgs($job)
 
-Given a Helios::Job object, getJobArgs() returns a hashref representing the parsed job argument 
-XML.  It actually calls the Helios::Job object's parseArgs() method and returns its value.
+Given a Helios::Job object, getJobArgs() returns a hashref representing the 
+parsed job argument XML.  It actually calls the Helios::Job object's parseArgs()
+method and returns its value.
 
 =cut
 
@@ -702,11 +836,11 @@ sub getJobArgs {
 }
 
 
-=head1 JOB METHODS
+=head1 JOB COMPLETION METHODS
 
-These methods should be called in the subclass's run() method to mark a job as successfully 
-completed, failed, or failed permanently.  They actually call the appropriate methods of the given 
-Helios::Job object.
+These methods should be called in your Helios service class's run() method to 
+mark a job as successfully completed, failed, or failed permanently.  They 
+actually call the appropriate methods of the given Helios::Job object.
 
 =head2 completedJob($job)
 
@@ -768,6 +902,53 @@ sub burstJob {
 }
 
 
+=head1 SERVICE CLASS DEFINITION
+
+These are the basic methods that define your Helios service.  The run() method 
+is the only one required. 
+
+=head2 run()
+
+This is a default run method for class completeness.  You have to override it 
+in your own Helios service class. 
+
+=cut
+
+sub run {
+    throw Helios::Error::FatalNoRetry($_[0]->getJobType.': run() method not implemented!'); 
+}
+
+=head2 MaxRetries() and RetryInterval()
+
+These methods control how many times a job should be retried if it fails and 
+how long the system should wait before a retry is attempted.  If you don't 
+defined these, jobs will not be retried if they fail.   
+
+=cut
+
+sub MaxRetries { return undef; }
+sub RetryInterval { return undef; }
+
+=head2 JobClass()
+
+Defines which job class to instantiate the job as.  The default is Helios::Job, 
+which should be fine for most purposes.  If necessary, however, you can create 
+a subclass of Helios::Job and set your JobClass() method to return that 
+subclass's name.  The service's work() method will instantiate the job as an 
+instance of the class you specified rather than the base Helios::Job.
+
+NOTE:  Please remember that "jobs" in Helios are most often only used to convey 
+arguments to services, and usually only contain enough logic to properly parse 
+those arguments and mark jobs as completed.  It should be rare to need to 
+extend the Helios::Job object.  OTOH, if you are attempting to extend Helios 
+itself to provide new abilities and not just writing a normal Helios 
+application, you can use JobClass() to use your extended job class rather than 
+the default.  
+
+=cut
+
+sub JobClass { return undef; }
+
 
 1;
 __END__
@@ -776,20 +957,27 @@ __END__
 =head1 SEE ALSO
 
 L<helios.pl>, L<Helios::Job>, L<Helios::Error>, L<Helios::ConfigParam>, L<Helios::LogEntry>, 
-L<TheSchwartz>, L<XML::Simple>, L<Config::IniFiles>
+L<TheSchwartz>, L<XML::Simple>, L<Config::IniFiles>, L<Sys::Syslog>
 
 =head1 AUTHOR
 
-Andrew Johnson, E<lt>ajohnson at ittoolbox dotcomE<gt>
+Andrew Johnson, E<lt>lajandy at cpan dot orgE<gt>
 
 =head1 COPYRIGHT AND LICENSE
 
-Copyright (C) 2008 by CEB Toolbox, Inc.
+Copyright (C) 2008-9 by CEB Toolbox, Inc., except as noted.
 
 This library is free software; you can redistribute it and/or modify
 it under the same terms as Perl itself, either Perl version 5.8.0 or,
 at your option, any later version of Perl 5 you may have available.
 
+
+The logMsg() method is Copyright (C) 2009 by Andrew Johnson.
+
+The logMsg() method is free software; you can redistribute it and/or modify
+it under the same terms as Perl itself, either Perl version 5.8.0 or,
+at your option, any later version of Perl 5 you may have available.
+
 =head1 WARRANTY
 
 This software comes with no warranty of any kind.
